@@ -94,6 +94,18 @@ class EdgeModel(ABC):
     def loss_perc(self):
         pass
 
+    def coords(self):
+        x = 0
+        y = 0
+        num_coords = 0
+        for node, _, _ in self.nodes:
+            coords = node.coords()
+            if coords is not None:
+                num_coords += 1
+                x += coords["x"]
+                y += coords["y"]
+        return {"x": x / num_coords, "y": y / num_coords} if num_coords > 0 else None
+
 
 class TrafoEdge(EdgeModel):
     def values_as_dict(self):
@@ -336,12 +348,12 @@ class PipeEdge(EdgeModel):
         return self._network.pipe[NAME_KEY][self._id]
 
     def calc_pipe_heat_loss_perc(self):
-        C = 1.161667
+        C = 500
         ALPHA = 0.19
         AMBIENT_T = 285
         d = self._network.pipe[DIAMETER_KEY][self._id]
         l = self._network.pipe[LENGTH_KEY][self._id] * 1000
-        m = (d / 2) * 2 * math.pi * l
+        m = (d / 2) ** 2 * math.pi * l
         delta_t = abs(self.temp_start() - self.temp_end()) + 1
         heat_energy = m * C * delta_t
         delta_ambient = abs(AMBIENT_T - (self.temp_start() + self.temp_end()) / 2)
