@@ -15,6 +15,8 @@ from peext.node import (
     CHPNode,
     EmptyBusNode,
     EmptyJunctionNode,
+    BusNode,
+    JunctionNode,
     ExtGasGrid,
     ExtPowerGrid,
     G2PNode,
@@ -53,13 +55,18 @@ class MENetwork:
     """
 
     def __init__(
-        self, nodes=None, edges=None, multinet=None, history_container=None
+        self,
+        nodes=None,
+        virtual_nodes=None,
+        edges=None,
+        multinet=None,
+        history_container=None,
     ) -> None:
         self.__nodes = nodes
         self.__edges = edges
+        self.__virtual_nodes = virtual_nodes
         self.__multinet = multinet
         self.__history_container = history_container
-        pass
 
     def add_history_table(self, network_name, table_name, table):
         if network_name not in self.__history_container:
@@ -121,6 +128,36 @@ class MENetwork:
             all_nodes += self.__nodes[key]
         return all_nodes
 
+    @property
+    def nodes_as_dict(self):
+        """Return all nodes as dict mapped to their network names
+
+        Returns:
+            dict: network name to list of nodes
+        """
+        return self.__nodes
+
+    @property
+    def virtual_nodes(self):
+        """Return all nodes as list
+
+        Returns:
+            _type_: list (MESModel)
+        """
+        all_nodes = []
+        for key in self.__virtual_nodes.keys():
+            all_nodes += self.__virtual_nodes[key]
+        return all_nodes
+
+    @property
+    def virtual_nodes_as_dict(self):
+        """Return all virtual nodes as dict mapped to their network names
+
+        Returns:
+            dict: network name to list of virtual nodes
+        """
+        return self.__virtual_nodes
+
     def power_nodes(self):
         """Return all nodes in the power network
 
@@ -241,7 +278,64 @@ def get_bus_junc_res_data(node):
     return data_set
 
 
-def find_edges_controller(edges, controller):
+def find_edges_controller(
+    multinet, controller, me_edges_index, edge_cache, cross_empty_cache
+):
+    """Finds all edges connecting to a controller.
+
+    Args:
+        edges (_type_): all edges
+        controller (_type_): the controller object
+
+    Returns:
+        _type_: dict{network_name -> list[edge_data...]}
+    """
+    all_edges = {}
+    for network, type_id_tuple in controller.nodes.items():
+        c_type, id = type_id_tuple
+        new_edges = {}
+        if c_type == "heat_exchanger":
+            new_edges = merge_dict_with_lists(
+                find_edges(
+                    me_edges_index,
+                    id,
+                    multinet["nets"][network],
+                    network,
+                    c_type,
+                    "to_junction",
+                    edge_cache,
+                    cross_empty_cache,
+                ),
+                find_edges(
+                    me_edges_index,
+                    id,
+                    multinet["nets"][network],
+                    network,
+                    c_type,
+                    "from_junction",
+                    edge_cache,
+                    cross_empty_cache,
+                ),
+            )
+        else:
+            new_edges = find_edges(
+                me_edges_index,
+                id,
+                multinet["nets"][network],
+                network,
+                c_type,
+                "bus" if network == "power" else "junction",
+                edge_cache,
+                cross_empty_cache,
+            )
+        all_edges = merge_dict_with_lists(
+            all_edges,
+            new_edges,
+        )
+    return all_edges
+
+
+def find_edges_controller_legacy(edges, controller):
     """Finds all edges connecting to a controller.
 
     Args:
@@ -434,6 +528,10 @@ def is_empty_junction(multinet, junction_id, network_name):
 
 
 def merge_dict_with_lists(first, second):
+    if len(first) == 0:
+        return second
+    if len(second) == 0:
+        return first
     result = {**first, **second}
     for key, value in result.items():
         if key in first and key in second:
@@ -476,6 +574,7 @@ def from_panda_multinet(multinet):
         heat_net = multinet[NETS_ACCESS][HEAT_NET]
 
     nodes = {GAS_NET: [], POWER_NET: [], HEAT_NET: [], MULTI: []}
+    virtual_nodes = {GAS_NET: [], POWER_NET: [], HEAT_NET: []}
     edges = {GAS_NET: [], POWER_NET: [], HEAT_NET: [], MULTI: []}
     history_container = {}
 
@@ -551,29 +650,78 @@ def from_panda_multinet(multinet):
                     ),
                 )
 
+    edge_cache = {}
+    cross_empty_cache = {GAS_NET: {}, POWER_NET: {}, HEAT_NET: {}, MULTI: {}}
+
     # controller
     for i, row in multinet.controller.iterrows():
         controller = row.object
 
         if isinstance(controller, RegulatedP2GControlMultiEnergy):
             nodes[MULTI].append(
-                post(P2GNode(i, multinet, find_edges_controller(edges, controller)))
+                post(
+                    P2GNode(
+                        i,
+                        multinet,
+                        find_edges_controller(
+                            multinet,
+                            controller,
+                            me_edges_index,
+                            edge_cache,
+                            cross_empty_cache,
+                        ),
+                    )
+                )
             )
         if isinstance(controller, RegulatedG2PControlMultiEnergy):
             nodes[MULTI].append(
-                post(G2PNode(i, multinet, find_edges_controller(edges, controller)))
+                post(
+                    G2PNode(
+                        i,
+                        multinet,
+                        find_edges_controller(
+                            multinet,
+                            controller,
+                            me_edges_index,
+                            edge_cache,
+                            cross_empty_cache,
+                        ),
+                    )
+                )
             )
         if isinstance(controller, CHPControlMultiEnergy):
             nodes[MULTI].append(
-                post(CHPNode(i, multinet, find_edges_controller(edges, controller)))
+                post(
+                    CHPNode(
+                        i,
+                        multinet,
+                        find_edges_controller(
+                            multinet,
+                            controller,
+                            me_edges_index,
+                            edge_cache,
+                            cross_empty_cache,
+                        ),
+                    )
+                )
             )
         if isinstance(controller, RegulatedP2HControlMultiEnergy):
             nodes[MULTI].append(
-                post(P2HNode(i, multinet, find_edges_controller(edges, controller)))
+                post(
+                    P2HNode(
+                        i,
+                        multinet,
+                        find_edges_controller(
+                            multinet,
+                            controller,
+                            me_edges_index,
+                            edge_cache,
+                            cross_empty_cache,
+                        ),
+                    )
+                )
             )
 
-    edge_cache = {}
-    cross_empty_cache = {GAS_NET: {}, POWER_NET: {}, HEAT_NET: {}, MULTI: {}}
     # power
     if POWER_NET in multinet[NETS_ACCESS]:
         if "load" in power_net:
@@ -726,9 +874,25 @@ def from_panda_multinet(multinet):
                         )
                     )
                 )
+            else:
+                virtual_nodes[POWER_NET].append(
+                    BusNode(
+                        i,
+                        power_net,
+                        find_edges(
+                            me_edges_index,
+                            i,
+                            power_net,
+                            POWER_NET,
+                            "bus",
+                            "bus",
+                            edge_cache,
+                            cross_empty_cache,
+                        ),
+                    )
+                )
 
     # gas
-
     if GAS_NET in multinet[NETS_ACCESS]:
         if "sink" in gas_net:
             for i in gas_net.sink.index:
@@ -828,6 +992,23 @@ def from_panda_multinet(multinet):
                         )
                     )
                 )
+            else:
+                virtual_nodes[GAS_NET].append(
+                    JunctionNode(
+                        i,
+                        gas_net,
+                        find_edges(
+                            me_edges_index,
+                            i,
+                            gas_net,
+                            GAS_NET,
+                            "junction",
+                            "junction",
+                            edge_cache,
+                            cross_empty_cache,
+                        ),
+                    )
+                )
 
     # heat
     if HEAT_NET in multinet[NETS_ACCESS]:
@@ -963,6 +1144,7 @@ def from_panda_multinet(multinet):
                         cross_empty_cache, heat_net, HEAT_NET, "ext_grid", "junction", i
                     )
 
+        real_junction_nodes = set()
         for i in heat_net.junction.index:
             if is_empty_cross_fastlane(cross_empty_cache, i, HEAT_NET):
                 nodes[HEAT_NET].append(
@@ -983,7 +1165,28 @@ def from_panda_multinet(multinet):
                         )
                     )
                 )
+            else:
+                virtual_nodes[HEAT_NET].append(
+                    JunctionNode(
+                        i,
+                        heat_net,
+                        find_edges(
+                            me_edges_index,
+                            i,
+                            heat_net,
+                            HEAT_NET,
+                            "junction",
+                            "junction",
+                            edge_cache,
+                            cross_empty_cache,
+                        ),
+                    )
+                )
 
     return MENetwork(
-        nodes=nodes, edges=edges, multinet=multinet, history_container=history_container
+        nodes=nodes,
+        virtual_nodes=virtual_nodes,
+        edges=edges,
+        multinet=multinet,
+        history_container=history_container,
     )
