@@ -53,6 +53,9 @@ class RegulatableController(RegulatableMESModel):
         """
         return self._network.controller[CONTROLLER_OBJ_KEY][self._id]
 
+    def coords(self):
+        return self.controller.coords
+
 
 class CouplingPoint(RegulatableController):
     def loss_perc(self):
@@ -555,6 +558,12 @@ class SourceNode(RegulatableMESModel):
 class HeatExchangerNode(RegulatableMESModel):
     """Node representing a heat exchanger"""
 
+    def __init__(self, id, network, edges=None) -> None:
+        super().__init__(id, network, edges)
+
+        self._initial_q = self.qext_w()
+        self._scaling = 1
+
     def mdot_kg_per_s(self):
         """
         :return: Actual mass flow
@@ -568,7 +577,7 @@ class HeatExchangerNode(RegulatableMESModel):
         :return: q cap
         :rtype: float
         """
-        return self._network.heat_exchanger[QEXT_W][self._id]
+        return self._initial_q
 
     def qext_w(self):
         """Return qext_w
@@ -584,14 +593,17 @@ class HeatExchangerNode(RegulatableMESModel):
         :param factor: new regulation factor
         :type factor: float [0,1]
         """
-        pass
+        self._scaling = new_scaling
+        self._network.heat_exchanger.at[self._id, QEXT_W] = (
+            new_scaling * self._initial_q
+        )
 
     def regulation_factor(self) -> float:
         """
         :return: the regulation factor for the heat_exchanger
         :rtype: float [0,1]
         """
-        return 1
+        return self._scaling
 
     def name(self):
         """Return name of the heat_exchanger
@@ -677,7 +689,7 @@ class ExtPowerGrid(RegulatableMESModel):
         return self._network.res_ext_grid.loc[self._id]
 
 
-class EmptyBusNode(RegulatableMESModel):
+class BusNode(RegulatableMESModel):
     def regulate(self, _):
         pass
 
@@ -701,8 +713,15 @@ class EmptyBusNode(RegulatableMESModel):
     def values_as_dict(self):
         return self._network.res_bus.loc[self._id]
 
+    def coords(self):
+        return self._network.bus_geodata.loc[self._id]
 
-class EmptyJunctionNode(RegulatableMESModel):
+
+class EmptyBusNode(BusNode):
+    pass
+
+
+class JunctionNode(RegulatableMESModel):
     def regulate(self, _):
         pass
 
@@ -725,3 +744,10 @@ class EmptyJunctionNode(RegulatableMESModel):
 
     def values_as_dict(self):
         return self._network.res_junction.loc[self._id]
+
+    def coords(self):
+        return self._network.junction_geodata.loc[self._id]
+
+
+class EmptyJunctionNode(JunctionNode):
+    pass
