@@ -1423,14 +1423,17 @@ def generate_multi_network_based_on_simbench(
     )
 
 
+def get_simbench_net(simbench_id):
+    return simbench.get_simbench_net(simbench_id)
+
+
 def create_gas_net_for_power(power_net, gas_deployment_rate, source_sink_ratio=0.05):
     net_gas = ppipes.create_empty_network(name="gas", fluid="lgas")
 
     bus_index_to_junction_index = {}
     for i in power_net.bus.index:
-        junc_id = ppipes.create_junction(
-            net_gas, pn_bar=REF_BAR, tfluid_k=REF_TEMP
-        )
+        bus_coords = (power_net.bus_geodata["x"][i], power_net.bus_geodata["y"][i])
+        junc_id = ppipes.create_junction(net_gas, pn_bar=REF_BAR, tfluid_k=REF_TEMP, geodata=bus_coords)
         bus_index_to_junction_index[i] = junc_id
 
     for i, from_bus, to_bus in zip(
@@ -1492,11 +1495,12 @@ def create_heat_net_for_power(power_net, heat_deployment_rate):
 
     bus_index_to_junction_index = {}
     for i in power_net.bus.index:
+        bus_coords = (power_net.bus_geodata["x"][i], power_net.bus_geodata["y"][i])
         junc_index_first = ppipes.create_junction(
-            net_heat, pn_bar=REF_BAR, tfluid_k=REF_TEMP
+            net_heat, pn_bar=REF_BAR, tfluid_k=REF_TEMP, geodata=bus_coords
         )
         junc_index_second = ppipes.create_junction(
-            net_heat, pn_bar=REF_BAR, tfluid_k=REF_TEMP
+            net_heat, pn_bar=REF_BAR, tfluid_k=REF_TEMP, geodata=bus_coords
         )
         bus_index_to_junction_index[i] = junc_index_first
         bus_index_to_junction_index[
@@ -1587,10 +1591,36 @@ def create_heat_net_for_power(power_net, heat_deployment_rate):
     return net_heat
 
 
-def create_p2h_in_combined_generated_network(mn, net_power, net_heat, p2h_density):
+def calc_cp_prob(attribute, id, cp_density, density_function=None, node_loc_map=None):
+    if density_function is None:
+        prob = cp_density
+    else:
+        node_loc = random.random()
+        if node_loc_map is not None:
+            node_loc = [
+                value
+                for key, value in node_loc_map.items()
+                if (attribute + ":" + str(id)) in key
+            ][0]
+        prob = density_function(node_loc, node_loc_map.values())
+
+    return prob
+
+
+def create_p2h_in_combined_generated_network(
+    mn, net_power, net_heat, p2h_density, density_function=None, node_loc_map=None
+):
     if "heat_exchanger" in net_heat:
         for i in net_heat.heat_exchanger.index:
-            if random.random() < p2h_density:
+            prob = calc_cp_prob(
+                "heat_exchanger",
+                i,
+                p2h_density,
+                density_function=density_function,
+                node_loc_map=node_loc_map,
+            )
+            has_to_build = random.random() < prob
+            if has_to_build:
                 cross_index = net_heat.heat_exchanger["from_junction"][i]
                 load = net_power.load[net_power.load["bus"] == cross_index]
                 if len(load.index) > 0:
@@ -1604,11 +1634,24 @@ def create_p2h_in_combined_generated_network(mn, net_power, net_heat, p2h_densit
 
 
 def create_chp_in_combined_generated_network(
-    mn, net_power, net_heat, net_gas, chp_density
+    mn,
+    net_power,
+    net_heat,
+    net_gas,
+    chp_density,
+    density_function=None,
+    node_loc_map=None,
 ):
     if "sink" in net_gas:
         for i in net_gas.sink.index:
-            if random.random() < chp_density:
+            prob = calc_cp_prob(
+                "sink",
+                i,
+                chp_density,
+                density_function=density_function,
+                node_loc_map=node_loc_map,
+            )
+            if random.random() < prob:
                 cross_index = net_gas.sink["junction"][i]
                 sgen = net_power.sgen[net_power.sgen["bus"] == cross_index]
                 heat_ex = net_heat.heat_exchanger[
@@ -1620,17 +1663,31 @@ def create_chp_in_combined_generated_network(
                         sgen.index[0],
                         i,
                         heat_ex.index[0],
-                        0.7 + random.random() * 0.3,
+                        0.45 + random.random() * 0.3,
                         "CHP-{i}",
                         ambient_temperature=282.5,
                         element_type_power="sgen",
                     )
 
 
-def create_p2g_in_combined_generated_network(mn, net_power, net_gas, p2g_density):
+def create_p2g_in_combined_generated_network(
+    mn,
+    net_power,
+    net_gas,
+    p2g_density,
+    density_function=None,
+    node_loc_map=None,
+):
     if "source" in net_gas:
         for i in net_gas.source.index:
-            if random.random() < p2g_density:
+            prob = calc_cp_prob(
+                "source",
+                i,
+                p2g_density,
+                density_function=density_function,
+                node_loc_map=node_loc_map,
+            )
+            if random.random() < prob:
                 cross_index = net_gas.source["junction"][i]
                 load = net_power.load[net_power.load["bus"] == cross_index]
                 if len(load.index) > 0:
